@@ -2,7 +2,7 @@
 
 # MIT License
 #
-# Copyright (c) 2021 Roberto Chamorro / project contributors
+# Copyright (c) 2022 Roberto Chamorro / project contributors
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -35,12 +35,15 @@ import yaml
 import shutil
 import logging
 import argparse
+import multiprocessing as mp
 
 from requests_oauthlib import OAuth1
+from requests.structures import CaseInsensitiveDict
 
 from .i18n import _
 from . import logger
 from .__init__ import __version__
+from ._utils import process_parallel, Locker, process_archive
 
 
 class User(object):
@@ -68,176 +71,142 @@ class User(object):
     from ._processing import process_tweets
 
     from ._processing import _expand_urls
+    from ._processing import _replace_url
     from ._processing import _get_media_url
     from ._processing import _process_polls
-    from ._processing import _replace_nitter
     from ._processing import _download_media
     from ._processing import _replace_mentions
     from ._processing import _get_best_bitrate_video
 
     def __init__(self, user_cfg: dict, cfg: dict, base_path: str):
-        self.twitter_token = cfg["twitter_token"]
-        self.signature = ""
-        self.media_upload = False
-        self.support_account = None
-        # iterate attrs defined in config
-        for attribute in user_cfg:
-            self.__setattr__(attribute, user_cfg[attribute])
-        self.twitter_url = f"http://twitter.com/{self.twitter_username}"
-        try:
-            if not hasattr(self, "max_tweets"):
-                self.max_tweets = cfg["max_tweets"]
-        except (KeyError, AttributeError):
-            # Limit to 50 last tweets - just to make a bit easier and faster
-            # to process given how often it is pulled
-            self.max_tweets = 50
-            pass
-        try:
-            if not hasattr(self, "visibility"):
-                self.visibility = cfg["visibility"]
-        except (KeyError, AttributeError):
-            self.visibility = "unlisted"
-            pass
-        visibility_valid = ("public", "unlisted", "private", "direct")
-        if self.visibility not in visibility_valid:
-            raise KeyError(
-                _(
-                    "Visibility not supported! Values allowed are: public, "
-                    "unlisted, private and direct"
-                )
-            )
-        try:
-            if not hasattr(self, "sensitive"):
-                self.sensitive = cfg["sensitive"]
-        except (KeyError, AttributeError):
-            pass
-        try:
-            if not hasattr(self, "delay_post"):
-                self.delay_post = cfg["delay_post"]
-        except (KeyError, AttributeError):
-            self.delay_post = 0.5
-            pass
-        try:
-            if not hasattr(self, "hashtags"):
-                self.hashtags = cfg["hashtags"]
-        except (KeyError, AttributeError):
-            self.hashtags = []
-            pass
-        try:
-            if not hasattr(self, "tweet_ids"):
-                self.tweet_ids = cfg["tweet_ids"]
-        except (KeyError, AttributeError):
-            self.tweet_ids = []
-            pass
-        if hasattr(self, "rich_text"):
-            if self.rich_text:
-                self.content_type = "text/markdown"
-        try:
-            if not hasattr(self, "pleroma_base_url"):
-                self.pleroma_base_url = cfg["pleroma_base_url"]
-        except KeyError:
-            raise KeyError(
-                _(
-                    "No Pleroma URL defined in config! [pleroma_base_url]"
-                )
-            )
-        try:
-            if not hasattr(self, "twitter_base_url"):
-                self.twitter_base_url = cfg["twitter_base_url"]
-        except KeyError:
-            self.twitter_base_url = "https://api.twitter.com/1.1"
-            pass
-        try:
-            if not hasattr(self, "twitter_base_url_v2"):
-                self.twitter_base_url = cfg["twitter_base_url_v2"]
-        except KeyError:
-            self.twitter_base_url_v2 = "https://api.twitter.com/2"
-            pass
-        try:
-            if not hasattr(self, "nitter_base_url"):
-                self.nitter_base_url = cfg["nitter_base_url"]
-        except KeyError:
-            self.nitter_base_url = "https://nitter.net"
-            pass
-        if not hasattr(self, "nitter"):
-            try:
-                if cfg["nitter"]:
-                    self.twitter_url = (
-                        f"{self.nitter_base_url}/{self.twitter_username}"
-                    )
-            except KeyError:
-                pass
-        else:
-            if self.nitter:
-                self.twitter_url = (
-                    f"{self.nitter_base_url}/{self.twitter_username}"
-                )
-        try:
-            if not hasattr(self, "include_rts"):
-                self.include_rts = cfg["include_rts"]
-        except (KeyError, AttributeError):
-            self.include_rts = True
-            pass
-        try:
-            if not hasattr(self, "include_replies"):
-                self.include_rts = cfg["include_replies"]
-        except (KeyError, AttributeError):
-            self.include_replies = True
-            pass
-        self.profile_image_url = None
-        self.profile_banner_url = None
-        self.display_name = None
+        self.posts = None
+        self.tweets = None
         self.first_time = False
-        try:
-            self.fields = self.replace_vars_in_str(str(user_cfg["fields"]))
-            self.fields = eval(self.fields)
-        except KeyError:
-            self.fields = []
-        self.bio_text = self.replace_vars_in_str(str(user_cfg["bio_text"]))
+        self.display_name = {}
+        self.last_post_pleroma = None
+        self.profile_image_url = {}
+        self.profile_banner_url = {}
+        self.t_user_tweets = {}
+        self.twitter_ids = {}
+        valid_visibility = ("public", "unlisted", "private", "direct")
+        default_cfg_attributes = {
+            "twitter_base_url": "https://api.twitter.com/1.1",
+            "twitter_base_url_v2": "https://api.twitter.com/2",
+            "nitter_base_url": "https://nitter.net",
+            "pleroma_base_url": None,
+            "nitter": False,
+            "twitter_token": None,
+            "signature": False,
+            "media_upload": True,
+            "sensitive": False,
+            "max_tweets": 50,
+            "delay_post": 0.5,
+            "visibility": "unlisted",
+            "hashtags": [],
+            "tweet_ids": [],
+            "rich_text": False,
+            "twitter_bio": True,
+            "include_rts": True,
+            "include_replies": True,
+            "consumer_key": None,
+            "consumer_secret": None,
+            "access_token_key": None,
+            "access_token_secret": None,
+            "original_date": False,
+            "original_date_format": "%Y-%m-%d %H:%M",
+            "bio_text": "",
+            "keep_media_links": False,
+            "fields": [],
+            "invidious": False,
+            "invidious_base_url": "https://yewtu.be/",
+            "archive": None,
+        }
+        # iterate attrs defined in config
+        for attribute in default_cfg_attributes:
+            if attribute in cfg:
+                self.__setattr__(attribute, cfg[attribute])
+            if not hasattr(self, attribute):
+                self.__setattr__(attribute, default_cfg_attributes[attribute])
+        for user_attribute in user_cfg:
+            self.__setattr__(user_attribute, user_cfg[user_attribute])
+
+        t_users = self.twitter_username
+        t_users_list = isinstance(t_users, list)
+        t_users = [t_users] if not t_users_list else t_users
+        self.twitter_username = t_users
+
+        twitter_url = (
+            self.nitter_base_url if self.nitter else "http://twitter.com"
+        )
+
+        if self.rich_text:
+            self.content_type = "text/markdown"
+        if not self.pleroma_base_url:
+            raise KeyError(
+                _("No Pleroma URL defined in config! [pleroma_base_url]")
+            )
+        if self.visibility not in valid_visibility:
+            raise KeyError(
+                _("Visibility not supported! Values allowed are: {}").format(
+                    ", ".join(valid_visibility)
+                )
+            )
+
+        bio_text = self.replace_vars_in_str(str(self.bio_text))
+        self.bio_text = {"_generic_bio_text": bio_text}
+
         # Auth
         self.header_pleroma = {"Authorization": f"Bearer {self.pleroma_token}"}
         self.header_twitter = {"Authorization": f"Bearer {self.twitter_token}"}
-        try:
-            if all([
+
+        if all(
+            [
                 self.consumer_key,
                 self.consumer_secret,
                 self.access_token_key,
-                self.access_token_secret
-            ]):
-                self.auth = OAuth1(
-                    self.consumer_key,
-                    self.consumer_secret,
-                    self.access_token_key,
-                    self.access_token_secret
-                )
-        except AttributeError:
+                self.access_token_secret,
+            ]
+        ):
+            self.auth = OAuth1(
+                self.consumer_key,
+                self.consumer_secret,
+                self.access_token_key,
+                self.access_token_secret,
+            )
+        else:
+            self.auth = None
             logger.debug(
                 _(
                     "Some or all OAuth 1.0a tokens missing, "
                     "falling back to application-only authentication"
                 )
             )
-            self.auth = None
 
-        self.tweets = None
+        self.twitter_url = CaseInsensitiveDict()
+        for t_user in t_users:
+            self.twitter_url[t_user] = f"{twitter_url}/{t_user}"
         self.pinned_tweet_id = self._get_pinned_tweet_id()
-        self.last_post_pleroma = None
-        # Filesystem
-        # self.base_path = os.getcwd()
+        self.fields = self.replace_vars_in_str(str(self.fields))
+        self.fields = eval(self.fields)
         self.base_path = base_path
         self.users_path = os.path.join(self.base_path, "users")
-        self.users_path = os.path.join(self.base_path, "users")
-        self.user_path = os.path.join(self.users_path, self.twitter_username)
-        self.tweets_temp_path = os.path.join(self.user_path, "tweets")
-        self.avatar_path = os.path.join(self.user_path, "profile.jpg")
-        self.header_path = os.path.join(self.user_path, "banner.jpg")
+        self.tweets_temp_path = os.path.join(self.base_path, "tweets")
+        self.user_path = {}
+        self.avatar_path = {}
+        self.header_path = {}
+        for t_user in t_users:
+            self.user_path[t_user] = os.path.join(self.users_path, t_user)
+            t_path = self.user_path[t_user]
+            self.avatar_path[t_user] = os.path.join(t_path, "profile.jpg")
+            self.header_path[t_user] = os.path.join(t_path, "banner.jpg")
         os.makedirs(self.users_path, exist_ok=True)
-        os.makedirs(self.user_path, exist_ok=True)
         os.makedirs(self.tweets_temp_path, exist_ok=True)
+        for t_user in t_users:
+            os.makedirs(self.user_path[t_user], exist_ok=True)
         # Get Twitter info on instance creation
         self._get_twitter_info()
         self._get_instance_info()
-        self.posts = None
+
         return
 
 
@@ -318,6 +287,19 @@ def get_args(sysargs):
         help=(_("skips first run checks")),
     )
 
+    parser.add_argument(
+        "-a",
+        "--archive",
+        required=False,
+        action="store",
+        help=(
+            _(
+                "path of the Twitter archive file (zip) to use for posting "
+                "tweets."
+            )
+        ),
+    )
+
     parser.add_argument("--verbose", "-v", action="count", default=0)
 
     parser.add_argument(
@@ -347,43 +329,50 @@ def main():
             base_path, cfg_file = os.path.split(os.path.abspath(config_path))
         else:
             config_path = os.path.join(base_path, "config.yml")
-
+        tweets_temp_path = os.path.join(base_path, "tweets")
+        logger.info(_("config path: {}").format(config_path))
+        logger.info(_("tweets temp folder: {}").format(tweets_temp_path))
+        # TODO: Add config generator wizard if config file is not found?
+        #  create a minimal config asking the user for the values
         with open(config_path, "r") as stream:
             config = yaml.safe_load(stream)
         user_dict = config["users"]
         users_path = os.path.join(base_path, "users")
-        # TODO: Merge tweets of multiple accounts and order them by date
         for user_item in user_dict[:]:
             user_item["skip_pin"] = False
-            if isinstance(user_item["twitter_username"], list):
+            t_users = user_item["twitter_username"]
+            t_user_list = isinstance(t_users, list)
+            t_users = t_users if t_user_list else [t_users]
+            if len(t_users) > 1:
                 warn_msg = _(
                     "Multiple twitter users for one Fediverse account, "
                     "skipping profile and pinned tweet."
                 )
                 logger.warning(warn_msg)
                 user_item["skip_pin"] = True
-                for twitter_user in user_item["twitter_username"]:
-                    new_user = dict(user_item)
-                    new_user["twitter_username"] = twitter_user
-                    user_dict.append(new_user)
-                user_dict.remove(user_item)
 
         for user_item in user_dict:
             first_time = False
             logger.info("======================================")
             logger.info(
-                _('Processing user:\t{}').format(user_item["pleroma_username"])
+                _("Processing user:\t{}").format(user_item["pleroma_username"])
             )
-            user_path = os.path.join(users_path, user_item["twitter_username"])
+            t_users = user_item["twitter_username"]
+            t_user_list = isinstance(t_users, list)
+            t_users = t_users if t_user_list else [t_users]
+            for t_user in t_users:
+                user_path = os.path.join(users_path, t_user)
 
-            if not os.path.exists(user_path):
-                first_time_msg = _(
-                    "It seems like pleroma-bot is running for the "
-                    "first time for this user"
-                )
-                logger.info(first_time_msg)
-                first_time = True
+                if not os.path.exists(user_path):
+                    first_time_msg = _(
+                        "It seems like pleroma-bot is running for the "
+                        "first time for this Twitter user: {}"
+                    ).format(t_user)
+                    logger.info(first_time_msg)
+                    first_time = True
             user = User(user_item, config, base_path)
+            if args.archive:
+                user.archive = args.archive
             if first_time and not args.skipChecks:
                 user.first_time = True
             if (
@@ -399,8 +388,9 @@ def main():
                 tweets = {
                     "data": [],
                     "includes": {},
-                    "meta": {"result_count": len(user.tweet_ids)}
+                    "meta": {"result_count": len(user.tweet_ids)},
                 }
+                user.result_count = len(user.tweet_ids)
                 includes = ["users", "tweets", "media", "polls"]
                 for include in includes:
                     try:
@@ -425,11 +415,14 @@ def main():
                         tweets["includes"]["media"].append(media)
                     for poll in next_tweet["includes"]["polls"]:
                         tweets["includes"]["polls"].append(poll)
+            elif args.archive:
+                tweets = process_archive(args.archive, start_time=date_pleroma)
+                user.result_count = len(tweets["data"])
             else:
                 tweets = user.get_tweets(start_time=date_pleroma)
             logger.debug(f"tweets: \t {tweets}")
 
-            if 'meta' not in tweets:
+            if "meta" not in tweets:
                 error_msg = _(
                     "Unable to retrieve tweets. Is the account protected?"
                     " If so, you need to provide the following OAuth 1.0a"
@@ -438,14 +431,20 @@ def main():
                     "- access_token_secret"
                 )
                 logger.error(error_msg)
-
-            if tweets["meta"]["result_count"] > 0:
+            if user.result_count > 0:
                 logger.info(
-                    _("tweet count: \t {}").format(len(tweets['data']))
+                    _("tweets gathered: \t {}").format(len(tweets["data"]))
                 )
                 # Put oldest first to iterate them and post them in order
                 tweets["data"].reverse()
-                tweets_to_post = user.process_tweets(tweets)
+                cores = mp.cpu_count()
+                threads = round(cores / 2 if cores > 4 else 4)
+                tweets_to_post = process_parallel(tweets, user, threads)
+                logger.info(
+                    _("tweets to post: \t {}").format(
+                        len(tweets_to_post['data'])
+                    )
+                )
                 logger.debug(f"tweets_processed: \t {tweets_to_post['data']}")
                 tweet_counter = 0
                 for tweet in tweets_to_post["data"]:
@@ -454,7 +453,7 @@ def main():
                         f"({tweet_counter}/{len(tweets_to_post['data'])})"
                     )
                     user.post_pleroma(
-                        (tweet["id"], tweet["text"]),
+                        (tweet["id"], tweet["text"], tweet["created_at"]),
                         tweet["polls"],
                         tweet["possibly_sensitive"],
                     )
@@ -501,7 +500,8 @@ def init():
     f_handler.setFormatter(f_format)
     logger.addHandler(f_handler)
     if __name__ == "__main__":
-        sys.exit(main())
+        with Locker():
+            sys.exit(main())
 
 
 init()
